/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package org.globus.gridshib.gt.authorization.attributes.tests.echoservice;

import org.globus.wsrf.client.BaseClient;
import org.globus.gridshib.gt.authorization.attributes.tests.echoservice.generated.SecurityContextEchoPortType;
import org.globus.gridshib.gt.authorization.attributes.tests.echoservice.generated.SecurityContextEchoServiceAddressingLocator;
import org.globus.gridshib.gt.authorization.attributes.tests.echoservice.generated.EchoRequest;
import org.globus.wsrf.utils.FaultHelper;

import javax.xml.rpc.Stub;

public class SecurityContextEchoClient extends BaseClient {

    public static void main(String[] args) {
        SecurityContextEchoClient client = new SecurityContextEchoClient();

        try {
            client.parse(args);
        } catch (Exception e) {
            System.err.println("Error: " + e.getMessage());
            System.exit(COMMAND_LINE_ERROR);
        }

        SecurityContextEchoServiceAddressingLocator locator =
            new SecurityContextEchoServiceAddressingLocator();

        try {
            SecurityContextEchoPortType port =
                   locator.getSecurityContextEchoPortTypePort(client.getEPR());
            client.setOptions((Stub)port);
            String echoResponse = port.echo(new EchoRequest());
            System.out.println("---------\nResponse:\n---------\n");
            System.out.println(echoResponse + "\n");
        } catch(Exception e) {
            if (client.isDebugMode()) {
                System.err.println("Error: " + FaultHelper.getMessage(e));
                FaultHelper.printStackTrace(e);
            } else {
                System.err.println("Error: " + FaultHelper.getMessage(e));
            }
            System.exit(APPLICATION_ERROR);
        }

    }

}
