/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib;

import org.globus.gridshib.gt.authorization.PDP;
import org.globus.gridshib.gt.authorization.SAMLAttributePDPImpl;

/**
 * This PDP makes an access control decision based on
 * information from the user's security context.  If
 * attributes from the security context satisfy policy,
 * the request is permitted.
 *
 * @see org.globus.gridshib.BasePDP
 * @see org.globus.gridshib.gt.authorization.SAMLAttributePDPImpl
 */
public class SAMLAttributePDP extends BasePDP {

    /**
     * The instance of <code>SAMLAttributePDPImpl</code>
     * that backs this <code>SAMLAttributePDP</code> instance.
     */
    protected SAMLAttributePDPImpl impl = null;

    /**
     * Gets the instance of <code>SAMLAttributePDPImpl</code>
     * that backs this <code>SAMLAttributePDP</code> instance.
     * If no such instance exists (presumably because this is
     * the first time the method has been called), this method
     * creates a new instance of <code>SAMLAttributePDPImpl</code>.
     *
     * @return an instance of <code>SAMLAttributePDPImpl</code>
     */
    protected PDP pdpImpl() {
        if (this.impl == null) {
            this.impl = new SAMLAttributePDPImpl();
        }
        return this.impl;
    }
}
