/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

import java.util.HashMap;
import java.io.File;

import javax.security.auth.Subject;
import javax.xml.rpc.handler.MessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.common.mapper.GridShibEntityMapper;
import org.globus.gridshib.gt.security.GS4GTSecurityContext;
import org.globus.gridshib.gt.security.util.GridShibSecurityUtil;
import org.globus.gridshib.security.util.SAMLUtil;

public class SAMLAssertionPushPIPImpl extends BasePIP
                                   implements GridShibConfigParams {

    private static Log logger =
        LogFactory.getLog(SAMLAssertionPushPIPImpl.class.getName());

    private static final String INITIALIZATION_MSG =
        "SAMLAssertionPushPIPImpl initialization complete";

    public void initialize(HashMap configs, String name) throws Exception {

        logger.info("SAMLAssertionPushPIPImpl initializing in scope " +
                    name);

        if (configs == null) {
            String msg = "HashMap configs is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("configs = " + configs.toString());

        if (this.isInitialized(name)) {
            logger.info("SAMLAssertionPushPIPImpl previously initialized " +
                        "in scope " + name);
            logger.info(INITIALIZATION_MSG);
            return;
        }

        // get path to a file or directory of metadata:
        String mapPath = (String)configs.get(METADATA_PATH_KEY);
        if (mapPath == null) {
            String msg = "Config parameter (" + METADATA_PATH_KEY +
                         ") not found";
            logger.error(msg);
            throw new Exception(msg);
        } else {
            logger.info("Config parameter (" + METADATA_PATH_KEY +
                         ") found: " + mapPath);
        }

        // see bug #5569
        File dirname = new File(mapPath);
        if (!dirname.isAbsolute()) {
            String msg = "Path is not absolute: \"" + mapPath + "\"";
            logger.error(msg);
            throw new Exception(msg);
        }

        // register map path with the EntityMapper:
        if (GridShibEntityMapper.register(mapPath)) {
            logger.info("Config parameter (" + METADATA_PATH_KEY +
                         ") successfully registered");
        } else {
            logger.warn("Config parameter (" + METADATA_PATH_KEY +
                        ") not registered");
        }

        // init security framework to use GS4GTSecurityContext impl:
        GS4GTSecurityContext.init();

        super.initialize(configs, name);
        logger.info(INITIALIZATION_MSG);
    }

    public void collectAttributes(Subject peerSubject,
                                  MessageContext context) throws Exception {

        if (peerSubject == null) {
            String msg = "Subject is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("peerSubject = " + peerSubject.toString());
        if (context == null) {
            String msg = "MessageContext is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("context = " + context.toString());

        String peerIdentity =
            GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
        logger.info("SAMLAssertionPushPIPImpl collecting attributes " +
                    "for identity " + peerIdentity);

        //SAMLUtil.traverseCertChain(peerSubject);
        SAMLUtil.consumeSAMLAssertions(peerSubject);

        logger.info("SAMLAssertionPushPIPImpl attribute collection complete");

        /* could be useful in a unit test
        String xmlPath;
        InputStream in;
        SAMLSubjectAssertion assertion;

        // get an assertion for testing:
        xmlPath = "/temp/testassertion.xml";
        logger.debug("XML path: " + xmlPath);
        in = new FileInputStream(xmlPath);
        assertion = new SAMLSubjectAssertion(in);
        logger.debug("Using assertion: " + assertion.toString());
        processAssertion(assertion, peerSubject);

        // get an assertion:
        xmlPath = "/temp/assertion.xml";
        logger.debug("XML path: " + xmlPath);
        in = new FileInputStream(xmlPath);
        assertion = new SAMLSubjectAssertion(in);
        logger.debug("Using assertion: " + assertion.toString());
        processAssertion(assertion, peerSubject);

        // get another assertion:
        xmlPath = "/temp/subjectAssertion.xml";
        logger.debug("XML path: " + xmlPath);
        in = new FileInputStream(xmlPath);
        assertion = new SAMLSubjectAssertion(in);
        logger.debug("Using assertion: " + assertion.toString());
        processAssertion(assertion, peerSubject);
        */
    }
}
