/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

import org.globus.gsi.CertUtil;
import org.globus.gsi.GlobusCredential;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Set;

public class X509ExtensionClient {

    public static void main(String[] args) throws Exception {

        System.out.println("EEC retrieved directly:");
        String filename =
            System.getenv("TEMP") + "\\x509up_u_" + System.getenv("USERNAME");
        //String filename =
        //    "C:\\Documents and Settings\\tom scavo\\Local Settings\\Temp\\x509up_u_tom scavo";
        InputStream in =
            new BufferedInputStream(new FileInputStream(filename));
        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        while (in.available() > 0) {
            Certificate cert = cf.generateCertificate(in);
            if (cert instanceof X509Certificate) {
                System.out.println("X.509 certificate:");
                printCertExtensions((X509Certificate)cert);
                System.out.println();
            } else {
                System.out.println("Certificate is not X.509 cert");
            }
        }
        in.close();

        GlobusCredential defaultCredential =
            GlobusCredential.getDefaultCredential();

        if (defaultCredential == null) {
            System.err.println("Error, no default credential");
            System.exit(-1);
        }

        X509Certificate[] certs = defaultCredential.getCertificateChain();

        String dn = defaultCredential.getIdentity();

        int eecIndex = -1;
        for (int i = 0; i < certs.length; i++) {
            String aDN = CertUtil.toGlobusID(certs[i].getSubjectDN());
            if (dn.equalsIgnoreCase(aDN)) {
                eecIndex = i;
                break;
            }
        }

        if (eecIndex == -1) {
            System.err.println("Error finding EEC (?)");
            System.exit(-1);
        }

        X509Certificate userx509 = certs[eecIndex];
        System.out.println("EEC retrieved with Globus methods:");
        printCertExtensions(userx509);
    }

    private static void printCertExtensions(X509Certificate cert) {
        System.out.println(cert.toString());

        System.out.println();

        System.out.println("Checking for critical extensions...");
        Set critSet = cert.getCriticalExtensionOIDs();
        if (critSet == null) {
            System.out.println("No extensions");
        } else if (critSet.isEmpty()) {
            System.out.println("No critical extensions");
        } else {
            System.out.println("Critical extension OIDs: " +
                               critSet.toString());
        }

        System.out.println();

        System.out.println("Checking for non-critical extensions...");
        Set nonCritSet = cert.getNonCriticalExtensionOIDs();
        if (nonCritSet == null) {
            System.out.println("No extensions");
        } else if (nonCritSet.isEmpty()) {
            System.out.println("No non-critical extensions");
        } else {
            System.out.println("Non-critical extension OIDs: " +
                               nonCritSet.toString());
        }
    }
}
