/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.security;

import java.security.Principal;
import javax.security.auth.Subject;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.security.SecurityContextLogger;

/**
 * This class is used to log the complete security context.
 * The format of the output is optimized for automated
 * processing, not human readability.
 */
public class GS4GTSecurityContextLogger {

    static Log logger =
        LogFactory.getLog(GS4GTSecurityContextLogger.class.getName());

    /**
     * Logs the security context of the given subject
     * at the INFO level
     *
     * @param subject the authenticated subject
     */
    public static void log(Subject subject) {

        GS4GTSecurityContext secCtx =
            GS4GTSecurityContext.getGS4GTSecurityContext(subject);
        assert (secCtx != null);

        Principal principal = secCtx.getGlobusPrincipal();
        if (principal != null) {
            String msg = "Principal" +
                         " Name=" +
                           quote(principal.getName()) +
                         " Type=Globus";
            logger.info(msg);
        }

        // TODO: move this code into SecurityContextLogger
        Principal[] samlPrincipals = secCtx.getSAMLPrincipals();
        assert (samlPrincipals != null);
        for (int i = 0; i < samlPrincipals.length; i++) {
            String msg = "Principal" +
                         " Name=" +
                           quote(samlPrincipals[i].getName()) +
                         " Type=SAML";
            logger.info(msg);
        }

        SecurityContextLogger.log(subject);

        Principal[] userNamePrincipals = secCtx.getUserNamePrincipals();
        assert (userNamePrincipals != null);
        for (int i = 0; i < userNamePrincipals.length; i++) {
            String msg = "Principal" +
                         " Name=" +
                           quote(userNamePrincipals[i].getName()) +
                         " Type=UserName";
            logger.info(msg);
        }
    }

    /*
     * redundant
     */
    private static String quote(String value) {

        assert (value != null);

        String s = value.replaceAll("\\\\", "\\\\");
        s = s.replaceAll("\"", "\\\"");
        return "\"" + s + "\"";
    }
}

