/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization.attributes.tests.aaservice;

import _0._1.shibboleth.mace.AttributeValueType;
import assertion._0._1.SAML.tc.names.oasis.AssertionType;
import assertion._0._1.SAML.tc.names.oasis.AttributeStatementType;
import assertion._0._1.SAML.tc.names.oasis.AttributeType;
import assertion._0._1.SAML.tc.names.oasis.AudienceRestrictionConditionType;
import assertion._0._1.SAML.tc.names.oasis.ConditionsType;
import assertion._0._1.SAML.tc.names.oasis.SubjectType;
import org.apache.axis.types.URI;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.samlAttributeRequest.generated.SAMLAttributeRequestPortType;
import org.globus.gridshib.gt.authorization.attributes.tests.AttributeTestConstants;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.NoAuthRequest;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.Policy;
import org.globus.gridshib.gt.authorization.ShibbolethConstants;
import protocol._0._1.SAML.tc.names.oasis.AttributeQueryType;
import protocol._0._1.SAML.tc.names.oasis.RequestType;
import protocol._0._1.SAML.tc.names.oasis.ResponseType;
import protocol._0._1.SAML.tc.names.oasis.StatusType;
import protocol._0._1.SAML.tc.names.oasis.StatusCodeType;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.math.BigInteger;
import java.rmi.RemoteException;
import java.util.Calendar;

public class TestAAService implements SAMLAttributeRequestPortType {

    static Log logger =
        LogFactory.getLog(TestAAService.class.getName());

    public ResponseType SAMLAttributeRequest(RequestType request) throws RemoteException {

        logger.debug("\n\n       ********************" +
                       "\n         INVOKING TEST AA  " +
                       "\n       ********************\n");

        ResponseType response = new ResponseType();

        String reqid = request.getRequestID();
        if (reqid == null) {
            throw new RemoteException("required attribute RequestID is absent");
        }
        response.setInResponseTo(reqid);
        response.setMajorVersion(BigInteger.ONE);
        response.setMinorVersion(BigInteger.ONE);
        response.setIssueInstant(Calendar.getInstance());

        AttributeQueryType attrQuery = request.getAttributeQuery();

        if (attrQuery == null) {
            throw new RemoteException("required element AttributeQuery is absent");
        }

        SubjectType subject = attrQuery.getSubject();

        if (logger.isDebugEnabled()) {
            logger.debug("\nReceived:\n" +
                "\n\tResource       = " + attrQuery.getResource() +
                "\n\tNameIdentifier = " + subject.getNameIdentifier().get_value() +
                "\n\tNameQualifier  = " + subject.getNameIdentifier().getNameQualifier() + "\n");
        }

        ResponseType resp = constructResponse(subject, request, response, attrQuery);

        logger.debug("\n\n       ******************" +
                       "\n         ENDING TEST AA  " +
                       "\n       ******************\n");
        return resp;
    }

    private ResponseType constructResponse(
            SubjectType subject,
            RequestType request,
            ResponseType response,
            AttributeQueryType attrQuery)
            throws RemoteException {
        NoAuthRequest control = null;
        try {
            InitialContext context = new InitialContext();
            logger.debug(org.globus.wsrf.Constants.JNDI_BASE_NAME +
                AttributeTestConstants.TEST_ATTRIBUTE_AUTHORITY +
                AttributeTestConstants.CONTROL_NAME);
             control = (NoAuthRequest)context.
                lookup(org.globus.wsrf.Constants.JNDI_BASE_NAME +
                AttributeTestConstants.TEST_ATTRIBUTE_AUTHORITY +
                AttributeTestConstants.CONTROL_NAME);
        } catch (NamingException e) {
            throw new RemoteException("cannot use without a test service" +
                    " that binds the control object into JNDI: ",e);
        }

        if (control.isAaError()) {
            constructError(response);
        } else {
            response.setAssertion(findAttributes(
                    subject,
                    getResource(control,attrQuery.getResource()),
                    control));
            response.setStatus(new StatusType(
                new StatusCodeType(null, ShibbolethConstants.SAMLP_1_0_SUCCESS),
                null,
                null));
        }

        return response;
    }

    private void constructError(ResponseType response) {
        // The shib AA does not return any information about the error
        StatusType status = new StatusType();
        status.setStatusCode(
             new StatusCodeType(null,ShibbolethConstants.SAMLP_1_0_RESPONDER));
        status.setStatusMessage("General error processing request.");
        response.setStatus(status);

    }

    private AssertionType[] findAttributes(SubjectType subject,
                                        URI resource,
                                        NoAuthRequest control) {

        AttributeStatementType attrStatement = new AttributeStatementType();
        attrStatement.setSubject(subject);
        AssertionType assertion;
        assertion = new AssertionType();

        if (!control.isAaEmptyAssertion()) {

            Policy okpolicy = control.getPolicy();

            if (okpolicy.getPolicyEntries() != null &&
                okpolicy.getPolicyEntries().getListOfAttributes() != null) {

                AttributeType[] attrs =
                        convertAttributes(okpolicy.getPolicyEntries()
                                      .getListOfAttributes().getAttribute());

                attrStatement.setAttribute(getAttrs(attrs,control));
                assertion.setAttributeStatement(attrStatement);
            }
        }
        assertion.setAssertionID("bogus-assertion-ID");
        assertion.setIssueInstant(Calendar.getInstance());
        assertion.setIssuer(getIssuer(control));
        assertion.setMajorVersion(BigInteger.ONE);
        assertion.setMinorVersion(BigInteger.ONE);
        URI[] audience = {resource};
        assertion.setConditions(getConditions(control,audience));

        AssertionType[] assertions = new AssertionType[1];
        assertions[0] = assertion;
        return assertions;
    }

    private AttributeType[] getAttrs(AttributeType[] attrs, NoAuthRequest control) {
        if (control.isAaMungeAttributes()) {
            for (int i = 0; i < attrs.length; i++) {
                attrs[i].setAttributeName(reverse(attrs[i].getAttributeName()));
                try {
                    attrs[i].setAttributeNamespace(
                            new URI(reverse(attrs[i].getAttributeNamespace().toString())));
                } catch (URI.MalformedURIException e) {
                    //just leave it
                    logger.debug("couldn't reverse uri");
                }
            }
            return attrs;
        } else {
            return attrs;
        }
    }


    private String getIssuer(NoAuthRequest control) {
        if (control.isAaMungeIssuer()) {
            return reverse(control.getPipIdPProviderId());
        } else {
            return control.getPipIdPProviderId();
        }
    }

    private URI getResource(NoAuthRequest control, URI queryresource) {
        if (control.isAaMungeResource()) {
            try {
                queryresource.setHost(reverse(queryresource.getHost()));
                return queryresource;
            } catch (Exception e) {
                try {
                    return new URI("https://blah.blah.com/blahblah/blah123");
                } catch (URI.MalformedURIException e1) {
                    return queryresource; // will cause test to fail
                }
            }
        } else {
            return queryresource;
        }
    }

    // for munging
    // won't work on palindromes ;-)
    private String reverse(String in) {
        return new StringBuffer(in).reverse().toString();
    }

    //man oh man
    private AttributeType[] convertAttributes(org.globus.gridshib.gt.authorization.attributes.tests.aaservice.generated.AttributeType[] oldattributes) {
        AttributeType[] newattributes = null;
        if (oldattributes != null) {
            newattributes = new AttributeType[oldattributes.length];
        } else {
            return newattributes = new AttributeType[0];
        }

        for (int i = 0; i < oldattributes.length; i++) {
            String oldname = oldattributes[i].getAttributeName();
            URI oldnamespace = oldattributes[i].getAttributeNamespace();
            org.globus.gridshib.gt.authorization.
                    attributes.tests.aaservice.generated.
                    AttributeValueType[] oldvals = oldattributes[i].getAttributeValue();
            AttributeValueType[] newvals = new AttributeValueType[oldvals.length];
            for (int j = 0; j < oldvals.length; j++) {
                newvals[j] = new AttributeValueType();
                newvals[j].setScope(oldvals[j].getScope());
                newvals[j].set_any(oldvals[j].get_any());
            }
            newattributes[i] = new AttributeType();
            newattributes[i].setAttributeValue(newvals);
            newattributes[i].setAttributeName(oldname);
            newattributes[i].setAttributeNamespace(oldnamespace);
        }

        return newattributes;
    }

    /* <Conditions NotBefore="2005-04-26T16:18:36.121Z"
        NotOnOrAfter="2005-04-26T16:48:36.121Z">
          <AudienceRestrictionCondition>
            <Audience>urn:mace:inqueue:example.org</Audience>
          </AudienceRestrictionCondition>
        </Conditions> */
    private ConditionsType getConditions(NoAuthRequest control, URI[] audience) {
        ConditionsType conditions = new ConditionsType();
        AudienceRestrictionConditionType restr = new AudienceRestrictionConditionType(audience);
        conditions.setAudienceRestrictionCondition(restr);

        Calendar start = Calendar.getInstance();
        start.add(Calendar.MINUTE, control.getAaIssueTimeOffset());
        conditions.setNotBefore(start);
        Calendar onafter = (Calendar)start.clone();
        onafter.add(Calendar.MINUTE, control.getAaValidityMinutes());
        conditions.setNotOnOrAfter(onafter);
        return conditions;
    }

}
