/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package org.globus.gridshib.gt.authorization.attributes.tests.service;

import org.apache.axis.MessageContext;
import org.apache.axis.message.MessageElement;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.gridshib.GridShibPDP;
import org.globus.gridshib.gt.authorization.ShibbolethConstants;
import org.globus.gridshib.gt.authorization.ShibbolethPDPPolicy;
import org.globus.gridshib.gt.authorization.attributes.AttributeAuthority;
import org.globus.gridshib.gt.authorization.attributes.SAMLAttribute;
import org.globus.gridshib.gt.authorization.attributes.policy.AttributeMapping;
import org.globus.gridshib.gt.authorization.attributes.tests.AttributeTestConstants;
import org.globus.gridshib.gt.authorization.attributes.tests.aaservice.generated.AttributeType;
import org.globus.gridshib.gt.authorization.attributes.tests.aaservice.generated.AttributeValueType;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.CreateResourceResponse;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.GetValue;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.ListOfAttributes;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.ListOfUsernames;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.NoAuthRequest;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.NoAuthResponse;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.Policy;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.PolicyDNs;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.PolicyEntries;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.SetValueResponse;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.ShibbolethTestPortType;
import org.globus.security.gridmap.GridMap;
import org.globus.wsrf.config.ConfigException;
import org.globus.wsrf.impl.security.authorization.ServicePropertiesPDPConfig;
import org.globus.wsrf.impl.security.descriptor.ServiceSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ServiceSecurityDescriptor;
import org.globus.wsrf.impl.security.util.PDPUtils;
import org.globus.wsrf.jndi.JNDIUtils;
import org.globus.wsrf.security.SecurityManager;
import org.globus.wsrf.utils.ContextUtils;
import org.oasis.wsn.GetCurrentMessage;
import org.oasis.wsn.GetCurrentMessageResponse;
import org.oasis.wsn.InvalidTopicExpressionFaultType;
import org.oasis.wsn.NoCurrentMessageOnTopicFaultType;
import org.oasis.wsn.ResourceUnknownFaultType;
import org.oasis.wsn.Subscribe;
import org.oasis.wsn.SubscribeCreationFailedFaultType;
import org.oasis.wsn.SubscribeResponse;
import org.oasis.wsn.TopicNotSupportedFaultType;
import org.oasis.wsn.TopicPathDialectUnknownFaultType;
import org.oasis.wsrf.properties.GetResourcePropertyResponse;
import org.oasis.wsrf.properties.InvalidResourcePropertyQNameFaultType;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.xml.namespace.QName;
import java.io.File;
import java.rmi.RemoteException;
import java.util.Vector;

/**
 * This service has an unauthorized operation whose parameter provides
 * the desired attribute authority and other information.
 * This is the GT4.0.x compatible version of this class.  See the build
 * script and var subdirectory of the test build directory.
 *
 * Please leave the DIFFERENT tags and surrounding whitespace, it helps
 * to anchor two-pane diff viewers for maintenance.
 *
 */
public class ShibbolethTestService implements ShibbolethTestPortType {

    static Log logger = LogFactory.getLog(ShibbolethTestService.class.getName());
    Context jndicontext = null;

    public ShibbolethTestService() throws Exception {
        Context context = JNDIUtils.initJNDI();
        context.createSubcontext(org.globus.wsrf.Constants.JNDI_BASE_NAME +
                        AttributeTestConstants.TEST_ATTRIBUTE_AUTHORITY);
    }

    public NoAuthResponse noAuth(NoAuthRequest req) throws RemoteException {

        StringBuffer msg = new StringBuffer();
        msg.append("broadcast:" +
                     "\n\n   ***************************" +
                       "\n     SETTING UP TEST SERVICE " +
                       "\n   ***************************");

        if (req.getTestname() != null) {
            msg.append("\n\n   TEST: \"")
                .append(req.getTestname())
                .append("\"");
        }
        if (req.getTestdescription() != null) {
            msg.append(" ")
               .append(req.getTestdescription());
        }

        msg.append("\n");
        logger.info(msg.toString());

        org.globus.wsrf.impl.security.authentication.wssec.GSSConfig.init();

        String attrService = req.getAttributeService();
        String idpId = req.getPipIdPProviderId();
        String spId = req.getPipSPProviderId();
        //String attrServiceIdentity = req.getAttributeServiceIdentity();

        try {
            setControl(req);
        } catch (NamingException e) {
            throw new RemoteException("problem with JNDI (using JNDI in " +
                    "this test to communicate parameters to the test AA", e);
        }

        MessageContext msgCtx = MessageContext.getCurrentContext();
        String servicePath = ContextUtils.getTargetServicePath(msgCtx);
        logger.debug("This service path: " + servicePath);

        String scope = "prefix1";

        // DIFFERENT
        String pdpChain = scope + ":" + GridShibPDP.class.getName();
        ServicePropertiesPDPConfig config =
                new ServicePropertiesPDPConfig(msgCtx,
                                               servicePath,
                                               pdpChain);




        // DIFFERENT
        config.setProperty(scope,
                           ShibbolethConstants.ENABLE_ATTRIBUTE_QUERY_KEY,
                           "true");




        // DIFFERENT
        config.setProperty(scope,
                           ShibbolethConstants.IDP_PROVIDER_ID_KEY,
                           idpId);


        logger.debug("IdP entityID set to " + idpId);



        // DIFFERENT
        config.setProperty(scope,
                           ShibbolethConstants.AA_INSTANCE_KEY,
                           new AttributeAuthority(attrService,
                                                  idpId,
                                                  idpId,
                                                  null, null, null));



        logger.debug("Attribute service set to " + attrService);


        // DIFFERENT

        config.setProperty(scope,
                           ShibbolethConstants.SP_PROVIDER_ID_KEY,
                           spId);



        logger.debug("SP entityID set to " + spId);

        String etcdir = req.getEtcdir();

        File path = new File(etcdir);

        if (!path.exists()) {
            throw new RemoteException("Cannot resolve test service etc " +
                   "directory, path does not exist: '" + etcdir + "'");
        }

        String metadataDir = etcdir + File.separator + "idp-metadata";

        path = new File(metadataDir);

        if (!path.exists()) {
            throw new RemoteException("Cannot resolve test service " +
                 "metadata directory, path does not exist: '" + etcdir + "'");
        }


        // DIFFERENT
        config.setProperty(scope,
                           ShibbolethConstants.METADATA_PATH_KEY,
                           metadataDir);




        // DIFFERENT
        setPolicy(req.getPolicy(), scope, config);


        // DIFFERENT
        try {
            ServiceSecurityDescriptor desc = ServiceSecurityConfig
                .getSecurityDescriptor(servicePath);
            desc.setAuthzChain(
                PDPUtils.getServiceAuthzChain(
                    config,
                    AttributeTestConstants.SHIB_PROTECTED_SERVICE_PATH));
            ServiceSecurityConfig.setSecurityDescriptor(desc, servicePath);
        } catch (ConfigException exp) {
            throw new RemoteException("", exp);
        }


        logger.debug("broadcast:" +
                     "\n\n   *****************************" +
                       "\n     ENDING TEST SERVICE SETUP  " +
                       "\n   *****************************\n");
        return null;
    }

    private void setPolicy(Policy policy,
                           String interceptorscope,
                           ServicePropertiesPDPConfig config)
                    throws RemoteException {

        PolicyEntries policyEntries = policy.getPolicyEntries();
        ListOfAttributes loa = policyEntries.getListOfAttributes();
        ListOfUsernames lou = policyEntries.getListOfUsernames();
        PolicyDNs policyDNs = policy.getPolicyDNs();

        String[] dns = null;
        if (policyDNs != null) {
            dns = policyDNs.getDN();
        }

        if (dns != null) {
            logger.debug("Policy DNs present, length " + dns.length);
        } else {
            logger.debug("No policy DNs");
        }

        String[] usernames = null;
        if (lou != null) {
            usernames = lou.getUsername();
        }
        if (usernames != null) {
            logger.debug("Usernames present, length " + usernames.length);
        } else {
            logger.debug("No usernames");
        }

        Vector newattrs = null;
        AttributeType[] attrs = null;
        if (loa != null) {
            attrs = loa.getAttribute();
        }
        if (attrs != null) {
            logger.debug("Attributes present, length " + attrs.length);
        } else {
            logger.debug("No attributes");
        }

        if (attrs != null) {
            newattrs = new Vector(attrs.length);
            for (int i = 0; i < attrs.length; i++) {
                AttributeType attr = attrs[i];
                String name = attr.getAttributeName();
                String namespace = null;
                if (attr.getAttributeNamespace() != null) {
                    namespace = attr.getAttributeNamespace().toString();
                }
                AttributeValueType[] oldvals = attr.getAttributeValue();
                String[] vals = new String[oldvals.length];
                for (int j = 0; j < oldvals.length; j++) {
                    String scope = oldvals[j].getScope();
                    MessageElement[] elms = oldvals[j].get_any();
                    String value = elms[0].getValue();
                    if (scope != null) {
                        value = value + "@" + scope;
                    }
                    vals[j] = value;
                }
                try {
                    // not testing issuer yet
                    newattrs.add(new SAMLAttribute(name,namespace,vals,
                                                   null,null,null));
                } catch (Exception e) {
                    throw new RemoteException("problem converting policy: ",e);
                }
            }
        }

        ShibbolethPDPPolicy attrPolicy =
                new ShibbolethPDPPolicy(new AttributeMapping[] {
                                  new AttributeMapping(newattrs, usernames) });

        config.setProperty(interceptorscope,
                   ShibbolethConstants.SHIB_PDP_POLICY_KEY,
                   attrPolicy);


        // as of 0.5, new way of doing in-memory DN policy, putting DNs in
        // ShibbolethPDPPolicy is deprecated

        if (dns != null) {

            GridMap gridmap = new GridMap();

            for (int i = 0; i < dns.length; i++) {
                gridmap.map(dns[i], "bogusAccount");
            }

            config.setProperty(interceptorscope,
                   ShibbolethConstants.CONSULT_GRIDMAP_KEY,
                   "true");

            config.setProperty(interceptorscope,
                   ShibbolethConstants.DEFAULT_GRIDMAP,
                   gridmap);
        }
    }

    private void setControl(NoAuthRequest req) throws NamingException {
        if (jndicontext == null) {
            jndicontext = new InitialContext();
            jndicontext.bind(org.globus.wsrf.Constants.JNDI_BASE_NAME +
                    AttributeTestConstants.TEST_ATTRIBUTE_AUTHORITY +
                    AttributeTestConstants.CONTROL_NAME, req);
        } else {
            jndicontext.rebind(org.globus.wsrf.Constants.JNDI_BASE_NAME +
                    AttributeTestConstants.TEST_ATTRIBUTE_AUTHORITY +
                    AttributeTestConstants.CONTROL_NAME, req);
        }

    }

    public CreateResourceResponse createResource(boolean parameters) throws RemoteException {

        logger.debug("broadcast:" +
                     "\n\n   *********************************" +
                       "\n     TEST SERVICE createResource() " +
                       "\n   *********************************\n");

        logger.debug("createResource() does nothing");

        return null;  //
    }

    public SetValueResponse setValue(int parameters) throws RemoteException {
        logger.debug("setValue called");

        SecurityManager manager = SecurityManager.getManager();
        String[] userNames = manager.getLocalUsernames();

        if (parameters == AttributeTestConstants.USERNAME_TEST_CODE) {

            if (userNames == null) {
                logger.error("user names are null");
                throw new RemoteException("user names are null");
            }

            int size = userNames.length;
            if ( size != AttributeTestConstants.USERNAME_LIST.length) {
                logger.error("username list size does not match");
                throw new RemoteException("username list size does not match");
            }



            for (int i=0; i<size; i++) {
                boolean match = false;
                for (int j=0; j<size; j++) {
                    if (AttributeTestConstants.USERNAME_LIST[i]
                        .equals(userNames[j])) {
                        match  = true;
                        break;
                    }
                }
                if (!match) {
                logger.error("username " + AttributeTestConstants
                             .USERNAME_LIST[i] + " does not exist");
                    throw new RemoteException("username "
                                              + AttributeTestConstants
                                              .USERNAME_LIST[i]
                                              + " does not exist");
                }
            }
        } else {
            if (userNames != null) {
                logger.error("User names is not null");
                for (int i=0; i<userNames.length; i++) {
                    logger.error("Offending username " + i + " "
                                 + userNames[i]);
                }
                throw new RemoteException("User names is set to something!");
            }
        }
        return null;  //
    }

    public int getValue(GetValue parameters) throws RemoteException {
        logger.debug("getValue");
        return 0;  //
    }

    public GetResourcePropertyResponse getResourceProperty(QName getResourcePropertyRequest) throws RemoteException, InvalidResourcePropertyQNameFaultType, org.oasis.wsrf.properties.ResourceUnknownFaultType {
        return null;  //
    }

    public SubscribeResponse subscribe(Subscribe subscribeRequest) throws RemoteException, TopicNotSupportedFaultType, InvalidTopicExpressionFaultType, SubscribeCreationFailedFaultType, ResourceUnknownFaultType, TopicPathDialectUnknownFaultType {
        return null;  //
    }

    public GetCurrentMessageResponse getCurrentMessage(GetCurrentMessage getCurrentMessageRequest) throws RemoteException, TopicNotSupportedFaultType, InvalidTopicExpressionFaultType, NoCurrentMessageOnTopicFaultType, ResourceUnknownFaultType {
        return null;  //
    }

}
