/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

import assertion._0._1.SAML.tc.names.oasis.AssertionType;
import assertion._0._1.SAML.tc.names.oasis.AttributeType;
import assertion._0._1.SAML.tc.names.oasis.SubjectType;
import org.apache.axis.client.Stub;
import org.apache.axis.types.URI;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.util.encoders.Hex;
import org.globus.axis.gsi.GSIConstants;
import org.globus.gridshib.gt.authorization.attributes.AttributeAuthority;
import org.globus.gridshib.gt.authorization.attributes.MessageInterface;
import org.globus.gridshib.gt.authorization.samlpiputils.CompatibleUtil;
import org.globus.gridshib.gt.authorization.samlpiputils.GridSP;
import org.globus.gridshib.gt.authorization.samlpiputils.MetadataUtil;
import org.globus.gridshib.gt.authorization.samlpiputils.SAMLutil;
import org.globus.gridshib.gt.security.util.GridShibSecurityUtil;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.gsi.jaas.JaasGssUtil;
import org.globus.samlAttributeRequest.generated.SAMLAttributeRequestPortType;
import org.globus.util.I18n;
import org.globus.wsrf.config.ConfigException;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.wsrf.impl.security.authorization.NoAuthorization;
import org.globus.wsrf.impl.security.authorization.IdentityAuthorization;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.gridshib.gt.authorization.attributes.AttributeAuthority;
import org.globus.gridshib.gt.authorization.attributes.MessageInterface;
import org.globus.gridshib.gt.authorization.samlpiputils.GridSP;
import org.globus.gridshib.gt.authorization.samlpiputils.MetadataUtil;
import org.globus.gridshib.gt.authorization.samlpiputils.SAMLutil;
import org.globus.gridshib.gt.security.GS4GTSecurityContext;
import org.globus.gridshib.gt.security.util.GridShibSecurityUtil;
import org.globus.wsrf.security.SecurityException;
import org.globus.wsrf.security.SecurityManager;
import org.globus.wsrf.utils.PerformanceLog;
import protocol._0._1.SAML.tc.names.oasis.AttributeQueryType;
import protocol._0._1.SAML.tc.names.oasis.RequestType;
import protocol._0._1.SAML.tc.names.oasis.ResponseType;
import protocol._0._1.SAML.tc.names.oasis.StatusCodeType;
import protocol._0._1.SAML.tc.names.oasis.StatusType;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;
import java.io.File;
import java.math.BigInteger;
import java.security.SecureRandom;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;

public class SAMLQueryPIPImpl extends BasePIP
                           implements ShibbolethConstants {

    private static I18n i18n =
        I18n.getI18n(
            MessageInterface.RESOURCE,
            SAMLQueryPIPImpl.class.getClassLoader()
        );

    static Log logger = LogFactory.getLog(SAMLQueryPIPImpl.class.getName());

    private static final String INITIALIZATION_MSG =
        "SAMLQueryPIPImpl initialization complete";

    private static PerformanceLog performanceLogger = null;

    private static SecureRandom random = new SecureRandom();

    private boolean isQueryEnabled = false;  // default
    private boolean cacheAttributes = false;
    private String cacheScope = null;

    // Object instead of Authorization because Authorization moved
    // out of the impl package in GT4.1+
    private Object authz = NoAuthorization.getInstance();
    private GridSP sp = null;
    private AttributeAuthority[] authorities = null;
    private AttributeAuthority configuredAA = null;

    boolean isAttributeQueryEnabled() {
        return this.isQueryEnabled;
    }

    public void initialize(HashMap configs, String name) throws Exception {

        logger.info("SAMLQueryPIPImpl initializing in scope " +
                    name);

        if (configs == null) {
            String msg = "HashMap configs is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("configs = " + configs.toString());

        if (this.isInitialized(name)) {
            logger.info("SAMLQueryPIPImpl previously initialized " +
                        "in scope " + name);
            logger.info(INITIALIZATION_MSG);
            return;
        }

        Object enableAttributeQuery = configs.get(ENABLE_ATTRIBUTE_QUERY_KEY);
        if (enableAttributeQuery != null) {
            String s = ((String)enableAttributeQuery).trim();
            if (s.equalsIgnoreCase("true")) {
                this.isQueryEnabled = true;
            } else if (s.equalsIgnoreCase("false")) {
                this.isQueryEnabled = false;
            } else {
                String msg = "Configuration parameter (" +
                             ENABLE_ATTRIBUTE_QUERY_KEY + ") is " +
                             "configured incorrectly: " + s;
                logger.error(msg);
                throw new RuntimeException(msg);
            }
        }
        if (this.isQueryEnabled) {
            logger.info("Attribute query is enabled");
            configs.put(ENABLE_ATTRIBUTE_QUERY_KEY, "true");
        } else {
            logger.info("Attribute query is disabled");
            configs.put(ENABLE_ATTRIBUTE_QUERY_KEY, "false");
            logger.info("SAMLQueryPIPImpl initialization complete");
            return;
        }

        /* If assertions are enabled, DEV will be set to true */
        boolean DEV = false;
        assert DEV = true;
        if (DEV) {
            performanceLogger = new PerformanceLog(
                           SAMLQueryPIPImpl.class.getName() + ".performance");
            performanceLogger.start(); //PIP initialization
        }

        /* ----------------  IdP config --------------- */

        // allow programmatic override
        Object o = configs.get(AA_INSTANCE_KEY);
        if (o != null) {
            logger.debug("config parameter (" + AA_INSTANCE_KEY +
                         ") found, overriding metadata if it exists");

            if (!(o instanceof AttributeAuthority)) {
                String err = i18n.getMessage("badAAInstance");
                logger.fatal(err);
                throw new Exception(err);
            }

            AttributeAuthority aa = (AttributeAuthority)o;
            this.authorities = new AttributeAuthority[]{aa};
        }

        /* From the deployer's point of view, the METADATA_PATH_KEY
         * config parameter is required.  If it's omitted, the
         * authorities member will be null, and it will throw an
         * exception.  However, it will also throw an exception if
         * there is metadata but no AttributeAuthority objects can
         * be obtained from that metadata.  So in that sense,
         * METADATA_PATH_KEY is a throw-away config parameter.
         */

        // directory of IdP metadata:
        String metadataDir = (String)configs.get(METADATA_PATH_KEY);
        if (this.authorities == null && metadataDir != null) {

            logger.debug("metadata directory found: " + metadataDir);

            // other code uses loadMetadataDir, doing the bug #5569 check here
            File dirname = new File(metadataDir);
            if (!dirname.isAbsolute()) {
                String err = i18n.getMessage("pathNotAbsolute", metadataDir);
                logger.fatal(err);
                throw new Exception(err);
            }

            // load is different than respect. This flag is here in case
            // we want to ease memory/time impact of parsing attributes
            // in.  Probably not, always true for now.  Config for
            // respecting the attributes is examined in GS4GT.PullPDPImpl.
            boolean loadMDAttributes = true;

            try {
                this.authorities =
                    MetadataUtil.loadMetadataDir(metadataDir,
                                                 loadMDAttributes);
            } catch (ConfigException e) {
                throw e;
            }
        }

        if (this.authorities == null) {
            String err = i18n.getMessage("noValidMetadata");
            logger.fatal(err);
            throw new Exception(err);
        }
        logger.debug("loaded " + this.authorities.length + " authorities");

        // default IdP entityID used for Classic GridShib:
        String idpEntityID = (String)configs.get(IDP_PROVIDER_ID_KEY);
        if (idpEntityID == null) {
            logger.debug("config parameter (" + IDP_PROVIDER_ID_KEY +
                         ") not found");
        } else {
            logger.debug("config parameter (" + IDP_PROVIDER_ID_KEY +
                         ") found: " + idpEntityID);
            this.configuredAA =
                AttributeAuthority.findAA(idpEntityID, this.authorities);
            if (this.configuredAA == null) {
                logger.debug("configured AA not found in metadata");
            } else {
                logger.debug("configured AA found in metadata");
            }
        }

        /* --------------  Grid SP  --------------- */

        /* SPproviderId */

        String SPproviderId = (String)configs.get(SP_PROVIDER_ID_KEY);

        if (SPproviderId == null) {
            String err = i18n.getMessage("noSPproviderId");
            logger.fatal(err);
            throw new Exception(err);
        }

        /* resourceAttr, defaults to SPproviderId */

        URI resourceAttr;
        try {
            resourceAttr = new URI(SPproviderId);
        } catch (URI.MalformedURIException e) {
            String err = i18n.getMessage("badSPURI", SPproviderId);
            logger.debug(err, e);
            throw new Exception(err, e);
        }

        this.sp = new GridSP(resourceAttr, SPproviderId);



        /* ------------  Query Authorization  --------------- */


        /* authz, for the call to the AA.
           Defaults to NO authorization */

        /* authz flag could have been set at previous configuration
           (i.e., it is not always true that initialize is called
           only at object instantiation time) */
        this.authz = NoAuthorization.getInstance();
        try {
            String authzconfig = (String) configs.get(AUTHZ_HOST_KEY);

            if (authzconfig != null) {
                // TODO: accept "false" as well
                if (authzconfig.trim().equalsIgnoreCase("true")) {
                    logger.debug("using host based AA authorization");
                    this.authz = HostAuthorization.getInstance();
                } else {
                    logger.debug("host authz flag configured, but " +
                            "does not match 'true'");
                    String err = i18n.getMessage("badAuthzNotTrue");
                    logger.error(err); // fatal?
                    throw new Exception(err);
                }
            } else {
                logger.debug("no configuration for AA authorization "
                        + "method present, defaulting to none");
            }
        } catch (ClassCastException e) {
            logger.error("authz configuration is present but it is " +
                    "not a String, programmer error");
            // This is an exception because if an attempt was made to
            // configure authorization and it was done incorrectly,
            // defaulting to no authz would be wrong.
            String err = i18n.getMessage("badAuthz");
            logger.error(err); // fatal?
            throw new Exception(err);
        }

        String authzDebugString = "Not configured";
        if (this.authz != null) {
            if (this.authz instanceof HostAuthorization) {
                authzDebugString = "Host based.";
            }
        }



        /* -------------------  Caching  --------------- */


        /* Check if caching is desired */

        /* any error for the cache flag is just logged and
           caching is set to off */

        /* cache flag could have been set at previous configuration
           (i.e., it is not always true that initialize is called
           only at object instantiation time) */
        this.cacheAttributes = false;
        try {
            Object cacheflag = configs.get(CACHE_KEY);

            if (cacheflag == null) {
                logger.debug("no cache configuration, default is false");
            } else {
                String cachestr = (String) cacheflag;
                if (cachestr.trim().equalsIgnoreCase("true")) {
                    logger.debug("caching activated");
                    this.cacheAttributes = true;
                    /* every authz chain scope will have its own cache scope */
                    this.cacheScope = name;
                } else {
                    logger.debug("cache flag configured, but does " +
                            "not match 'true'");
                }
            }
        } catch (ClassCastException e) {
            logger.warn("caching configuration is present but it is " +
                    "not the proper class, programmer error");
        } catch (Exception e) {
            logger.warn(e.getMessage());
        }

        String cacheDebugString = "Not configured";
        if (this.cacheAttributes) {
            cacheDebugString = "Configured, scope: " + this.cacheScope;
        }

        if (DEV) {
            performanceLogger.stop("PIP initialization");
        }

        /* finalize */

        logger.info("\n\nShibboleth attribute query initialized:\n"
                  + "\n        SP Provider ID : " + this.sp.getSPproviderId()
                  + "\n  AA/IdP Authorization : " + authzDebugString
                  + "\n               Caching : " + cacheDebugString
                  + "\n  # supported AAs/IdPs : " + this.authorities.length
                  + "\n");


        if (logger.isDebugEnabled() && this.authorities.length > 0) {
            StringBuffer buf = new StringBuffer();

            buf.append("\n\nproviderIds:\n\n");

            for (int i = 0; i < this.authorities.length; i++) {
                buf.append("#").
                    append(i).
                    append(": ").
                    append(this.authorities[i].getAAProviderId()).
                    append("\n    AAUrl = ").
                    append(this.authorities[i].getAAURL()).
                    append("\n");
            }

            logger.debug(buf.toString());
        }

        super.initialize(configs, name);
        logger.info(INITIALIZATION_MSG);
    }

    public void collectAttributes(Subject peerSubject,
                                  MessageContext context) throws Exception {

        if (peerSubject == null) {
            String msg = "Subject is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("peerSubject = " + peerSubject.toString());
        if (context == null) {
            String msg = "MessageContext is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("context = " + context.toString());

        String peerIdentity =
            GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
        logger.info("SAMLQueryPIPImpl collecting attributes " +
                    "for identity " + peerIdentity);

        if (this.isQueryEnabled) {
            logger.info("Attribute query is enabled");
        } else {
            logger.info("Attribute query is disabled");
            return;
        }
        assert (this.isQueryEnabled);

        AttributeAuthority aa = null;
        SubjectType subject = null;

        // get security context:
        GS4GTSecurityContext secCtx =
            GS4GTSecurityContext.getGS4GTSecurityContext(peerSubject);
        assert (secCtx != null);

        /* The <code>currentIdentity</code> is an <code>Identity</code>
         * object used primarily for query purposes.  An application
         * may set the <code>currentIdentity</code> object explicitly
         * with the #setCurrentIdentity(Identity) method.  Barring that,
         * the <code>currentIdentity</code> is the first item in the
         * ordered list of SAML <code>Identity</code> objects associated
         * with this <code>SecurityContext</code>.
         */
        if (secCtx.hasCurrentSAMLIdentity()) {
            aa = SAMLutil.chooseAA(this.authorities, peerSubject);
            if (aa == null) {
                logger.warn("No AA in metadata to query, " +
                            "aborting attribute collection");
                return;
            }
            subject = SAMLutil.constructSubjectFromSAMLIdentity(peerSubject);
        } else {
            aa = this.configuredAA;
            if (aa == null) {
                logger.warn("No configured AA to query, " +
                            "aborting attribute collection");
                return;
            }
            subject = SAMLutil.constructSubjectFromX509Identity(peerSubject);
        }

        AssertionType[] assertions = queryAttributes(aa, subject, context);
        for (int i = 0; i < assertions.length; i++) {

            AttributeType[] attrs =
                        SAMLutil.parseAssertion(assertions[i],
                                                subject,
                                                this.sp,
                                                aa);

            // checked for proper aa.getAAProviderId() in parseAssertion()

            // checked these for nulls in parseAssertion()
            Calendar notBefore =
                    assertions[i].getConditions().getNotBefore();
            Calendar notOnOrAfter =
                    assertions[i].getConditions().getNotOnOrAfter();

            Collection attributes = new HashSet();
            for (int j = 0; j < attrs.length; j++) {

                org.globus.opensaml11.saml.SAMLAttribute a;
                a = SAMLutil.parseAttribute(attrs[j],
                                            notBefore,
                                            notOnOrAfter);
                attributes.add(a);
            }

            if (attributes != null) {
                secCtx.addSAMLAttributes(assertions[i].getAssertionID(),
                                         aa.getAAProviderId(),
                                         attributes.iterator());
            }
        }

        logger.info("SAMLQueryPIPImpl attribute collection complete");
    }

    private AssertionType[] queryAttributes(AttributeAuthority aa,
                                            SubjectType subject,
                                            MessageContext context)
                                     throws Exception {

        assert (aa != null);
        assert (subject != null);
        assert (context != null);

        RequestType requestType = new RequestType();
        requestType.setMajorVersion(getMajorVersion());
        requestType.setMinorVersion(getMinorVersion());
        requestType.setIssueInstant(Calendar.getInstance());
        requestType.setRequestID(getIdentifier());

        AttributeQueryType query = new AttributeQueryType();
        query.setResource(this.sp.getResourceURI());
        query.setSubject(subject);
        //query.setAttributeDesignator(getAttributeDesignatorTypes());
        requestType.setAttributeQuery(query);

        ResponseType responseType;
        try {
            SAMLAttributeRequestPortType attrPort =
                    CompatibleUtil.getPort(aa.getAAURL());

            if (aa.getAAURL().trim().startsWith("https")) {

                GlobusGSSCredentialImpl credential = getCredential(context);

                ((Stub)attrPort)._setProperty(GSIConstants.GSI_CREDENTIALS,
                        credential);
                ((Stub)attrPort)._setProperty(GSIConstants.GSI_TRANSPORT,
                        Constants.ENCRYPTION);
                ((Stub)attrPort)._setProperty(Constants.AUTHORIZATION,
                        this.authz);
            } else if (aa.getAAURL().trim().startsWith("http")) {
                //
            } else {
                String err =
                        i18n.getMessage("unsupportedScheme", aa.getAAURL());
                throw new Exception(err);
            }

            logger.debug("Sending request");

            responseType = attrPort.SAMLAttributeRequest(requestType);
        } catch (Exception e) {
            String err = i18n.getMessage("attrService", aa.getAAURL());
            logger.debug(err, e);
            throw new Exception(err, e);
        }

        logger.debug("Response received");

        if (responseType == null) {
            String err = "responseType null";
            logger.debug(err);
            throw new Exception(err);
        }


        // Assert that request id matches the InResponseTo
        String inResponseTo = responseType.getInResponseTo();

        if ((inResponseTo == null) ||
            (!inResponseTo.equals(requestType.getRequestID()))) {
            // inResponseTo is optional in schema
            String err = i18n.getMessage("badInResponseTo");
            logger.warn(err);
            // TODO: decide which - maybe null is OK,
            // but !equals gives exception?
            //throw new AttributeException(err);
        } else {
            logger.debug("InResponseTo OK");
        }

        StatusType status = responseType.getStatus();
        if (status == null) {
            // status is required in schema
            String err = i18n.getMessage("responseNoStatus");
            throw new Exception(err);
        } else {
            //schema allows chained statuses, we only look at first one
            StatusCodeType statuscode = status.getStatusCode();
            if (statuscode == null) {
                // statuscode is required in schema
                String err = i18n.getMessage("responseNoStatuscode");
                throw new Exception(err);
            }
            logger.debug("status code: " + statuscode.getValue().toString());

            QName statusValue =  statuscode.getValue();
            if (statusValue == null) {
                // statusvalue is required in schema
                String err = i18n.getMessage("responseNoStatusvalue");
                throw new Exception(err);
            }

            if (statusValue.equals(ShibbolethConstants.SAMLP_1_0_SUCCESS)
                || statusValue.equals(ShibbolethConstants.SAMLP_1_1_SUCCESS)) {

                logger.debug("samlp success");
            } else if (statuscode.getValue().equals(
                    ShibbolethConstants.SAMLP_1_0_RESPONDER)
                || statuscode.getValue().equals(
                    ShibbolethConstants.SAMLP_1_1_RESPONDER)) {

                logger.debug("response equals responder");

                logger.info("\n\n" +
                        "Received response from Attribute Authority:\n" +
                        status.getStatusMessage() + "\n");

                return null;
            }
        }

        logger.trace("Getting assertions");
        AssertionType[] assertions = responseType.getAssertion();

        if (assertions == null) {
            logger.warn("found no assertions");
        } else {
            if (logger.isDebugEnabled()) {
                if (assertions.length == 1)
                    logger.debug("found one assertion");
                else
                    logger.debug("found " + assertions.length + " assertions");
            }
            return assertions;
        }

        return null;
    }

    private GlobusGSSCredentialImpl getCredential(MessageContext context)
                                           throws SecurityException {

        org.apache.axis.MessageContext ctx =
            (org.apache.axis.MessageContext)context;
        SecurityManager manager = SecurityManager.getManager(ctx);
        Subject serviceSubject= manager.getServiceSubject();

        GlobusGSSCredentialImpl credential =
            (GlobusGSSCredentialImpl)JaasGssUtil.
                getCredential(serviceSubject);

        return credential;
    }

    private BigInteger getMinorVersion() {
        return BigInteger.ONE;
    }

    private BigInteger getMajorVersion() {
        return BigInteger.ONE;
    }

    private synchronized String getIdentifier() {
        byte[] buf=new byte[16];
        random.nextBytes(buf);
        return "_".concat(new String(Hex.encode(buf)));
    }

    public AttributeAuthority[] getAuthorities() {
        return authorities;
    }

}
