/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package org.globus.gridshib.gt.authorization;

import java.util.ArrayList;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.gt.authorization.attributes.policy.AttributeMapping;

public class ShibbolethPDPPolicy {

    static Log logger =
        LogFactory.getLog(ShibbolethPDPPolicy.class.getName());

    private ArrayList DNs = null; //todo: Set?
    private ArrayList attrMappings = null;

    // DNs in policy are deprecated

    public ShibbolethPDPPolicy(AttributeMapping[] _attrs) {
        this(null, _attrs);
    }

    public ShibbolethPDPPolicy(String[] _DNs, AttributeMapping[] _attrs) {
        this.DNs = new ArrayList();
        if (_DNs != null) {
            for (int i = 0; i < _DNs.length; i++) {
                this.DNs.add(_DNs[i]);
            }
        }
        this.attrMappings = new ArrayList();
        if (_attrs != null) {
            for (int i = 0; i < _attrs.length; i++) {
                this.attrMappings.add(_attrs[i]);
            }
        }
    }

    public AttributeMapping[] getAttrMappingsArray() {
        return (AttributeMapping[])
                this.attrMappings.toArray(new AttributeMapping[0]);
    }

    public Iterator getAttrMappings() {
        return this.attrMappings.iterator();
    }

    /**
     * @deprecated
     * @param DN The DN to add to policy.
     * @return true if policy was altered.
     */
    public boolean addDN(String DN) {
        if (!this.DNs.contains(DN)) {
            this.DNs.add(DN);
            return true;
        }
        return false;
    }

    /**
     * @deprecated
     * @param DN The DN to remove from policy.
     * @return true if policy was altered.
     */
    public boolean removeDN(String DN) {
        return this.DNs.remove(DN);
    }

    /**
     * @deprecated
     * @param _DNs The DNs to replace the current DN policy
     * @return true if policy was altered.
     */
    public boolean replaceDNs(String[] _DNs) {
        this.DNs = new ArrayList();
        if (_DNs != null) {
            for (int i = 0; i < _DNs.length; i++) {
                this.DNs.add(_DNs[i]);
            }
        }
        return true;
    }

    /**
     * @param attr The attribute to add to policy.
     * @return true if policy was altered.
     */
    public boolean addAttrMapping(AttributeMapping attr) {
        if (!this.attrMappings.contains(attr)) {
            this.attrMappings.add(attr);
            return true;
        }
        return false;
    }

    /**
     * @param attr The attribute to remove from policy.
     * @return true if policy was altered.
     */
    public boolean removeAttr(AttributeMapping attr) {
        return this.attrMappings.remove(attr);
    }


    /**
     * @param _attrs The attributes to replace the current attribute policy
     * @return true if policy was altered.
     */
    public boolean replaceAttrs(AttributeMapping[] _attrs) {
        this.attrMappings = new ArrayList();
        if (_attrs != null) {
            for (int i = 0; i < _attrs.length; i++) {
                this.attrMappings.add(_attrs[i]);
            }
        }
        return true;
    }

    /**
     * Delete all entries.
     */
    public void clearPolicy() {
        this.DNs.clear();
        this.attrMappings.clear();
    }


    public String toString() {
        return "ShibbolethPDPPolicy{" +
                "attrs=" + attrMappings +
                ", DNs=" + DNs +
                "}";
    }
}
