/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package org.globus.gridshib.gt.authorization;

import org.globus.gsi.GlobusCredential;
import org.globus.gsi.CertUtil;

import javax.security.auth.x500.X500Principal;
import java.security.cert.X509Certificate;

public class X500DNClient {
    public static void main(String[] args) throws Exception {
        GlobusCredential defaultCredential =
                GlobusCredential.getDefaultCredential();

        if (defaultCredential == null) {
            System.err.println("Error, no default credential");
            System.exit(-1);
        }

        X509Certificate certs[] = defaultCredential.getCertificateChain();

        String dn = defaultCredential.getIdentity();

        int eecIndex = -1;
        for (int i = 0; i < certs.length; i++) {
            String aDN = CertUtil.toGlobusID(certs[i].getSubjectDN());
            if (dn.equalsIgnoreCase(aDN)) {
                eecIndex = i;
                break;
            }
        }

        if (eecIndex == -1) {
            System.err.println("Error finding EEC (?)");
            System.exit(-1);
        }

        X509Certificate userx509 = certs[eecIndex];

        System.out.println("X.509 Subject DN (Globus):");
        System.out.println();
        System.out.println(dn);
        System.out.println();
        X500Principal subjectDN = userx509.getSubjectX500Principal();
        System.out.println("X.509 Subject DN (RFC 2253):");
        System.out.println();
        System.out.println(subjectDN.getName(X500Principal.RFC2253));
        System.out.println();
        System.out.println("X.509 Subject DN (Canonical RFC 2253):");
        System.out.println();
        System.out.println(subjectDN.getName(X500Principal.CANONICAL));
        System.out.println();
    }

}
