/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.security.util;

import org.globus.gsi.jaas.GlobusPrincipal;

import javax.security.auth.Subject;
import java.util.Set;
import java.util.Iterator;

/**
 * Utility class for generic, security related utilities that are either
 * not found elsewhere or whose implementations are GT4.0/4.1+ API specific
 * if they are in Globus Java core.
 */
public class GridShibSecurityUtil {

    /**
     * Because AuthUtil.getIdentity() is 4.0 specific and
     * CredentialUtil.getIdentity() is 4.1+ specific, copying
     * implementation into our code.
     *
     * @param subject Subject
     * @return peerIdentity identity string (DN), null if given subject is null or if no principal in subject
     */
    public static String getIdentityFromSubject(Subject subject) {
        if (subject == null) {
            return null;
        }

        Set principals = subject.getPrincipals(GlobusPrincipal.class);

        if ((principals == null) || principals.isEmpty()) {
            return null;
        }

        Iterator iter = principals.iterator();
        GlobusPrincipal principal = (GlobusPrincipal) iter.next();

        return principal.toString();
    }
}
