/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package org.globus.gridshib.gt.authorization.attributes.tests.service;

import org.globus.wsrf.security.SecureResource;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.TopicListAccessor;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.TopicList;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.Topic;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.wsrf.impl.security.SecurityTestResource;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.SimpleTopicList;
import org.globus.wsrf.impl.ResourcePropertyTopic;
import org.globus.wsrf.impl.SimpleResourceProperty;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.xml.namespace.QName;

public class ShibbolethTestResource implements SecureResource,
        Resource,
        ResourceProperties,
        TopicListAccessor {

    ResourceSecurityDescriptor desc = null;

    static Log logger =
        LogFactory.getLog(SecurityTestResource.class.getName());

    public static final QName RP_SET =
        new QName("http://shibboleth.test", "TestRP");

    public static final QName VALUE =
        new QName("http://shibboleth.test", "Value");

    private ResourcePropertySet propSet;
    private TopicList topicList;
    protected ResourceProperty value;

    public ShibbolethTestResource() {
        this.propSet = new SimpleResourcePropertySet(RP_SET);
        this.topicList = new SimpleTopicList(this);

        try {
            this.value = new ResourcePropertyTopic(
                new SimpleResourceProperty(VALUE));
            this.propSet.add(this.value);
            this.topicList.addTopic((Topic) this.value);
            this.value.add(new Integer(0));
        } catch (Exception exp) {
            throw new RuntimeException(exp.getMessage());
        }
    }

    public TopicList getTopicList() {
        return this.topicList;
    }

    public ResourcePropertySet getResourcePropertySet() {
        return this.propSet;
    }

    public void setValue(int value) {
        this.value.set(0, new Integer(value));
    }

    public int getValue() {
        return ((Integer) this.value.get(0)).intValue();
    }

    public ResourceSecurityDescriptor getSecurityDescriptor() {
        if (this.desc != null) {
            // no getAuthz() call in 4.1+
            //logger.debug("Authz at get is " + this.desc.getAuthz());
        } else {
            logger.debug("Authz is null");
        }
        return this.desc;
    }
}
